#!/usr/bin/python
#
# TUI for RHN Registration
# Copyright (c) 2000-2002 Red Hat, Inc.
#
# Author:
#       Adrian Likins <alikins@redhat.com>
#       Preston Brown <pbrown@redhat.com>

from socket import gethostname
from os import geteuid, chdir
import sys
sys.path.insert(0, "/usr/share/rhn/")
sys.path.insert(1,"/usr/share/rhn/up2date_client")

from rhpl.translate import _, N_

import snack

import signal

sys.path.append("/usr/share/rhn/up2date")

import rhnreg, hardware
import up2dateErrors
import up2dateUtils
import rpmUtils
import rpcServer

import config

cfg = config.initUp2dateConfig()
swcfg = cfg.swcfg


def FatalErrorAbort( errmsg):
    print 'Fatal Error: %s' %  errmsg ;
    sys.exit(1) 

def PrintMess( msg ):
    print ( '%s' %  msg )
    return 

    
class StartWindow:
    
    def __init__(self, tui):
       self.name = "StartWindow"
       self.tui = tui

       try:
	    self.rhnregwelcomeText = rhnreg.welcomeText() ;
       except up2dateErrors.CommunicationError, e:
            FatalErrorAbort( e.errmsg)

       rhnreg.getCaps()

    def saveResults(self):
        pass

    def run(self):
         return "next"
    

class PrivacyWindow:

    def __init__(self,  tui):
        self.name = "PrivacyWindow"

        self.tui = tui

        try:
	    self.rhnregprivacyText = rhnreg.privacyText() ;
        except up2dateErrors.CommunicationError, e:
            FatalErrorAbort( e.errmsg)


    def saveResults(self):
        pass

    def run(self):
        return "next"

class InfoWindow:

    def __init__(self, tui):

        self.name = "InfoWindow"

        self.tui = tui
        self.tui.alreadyRegistered = 0
        
	self.userNameEntry = swcfg.GetLogin()
	self.passwordEntry = swcfg.GetPassword()
	self.emailEntry = swcfg.GetEmail()


    def validateFields(self):
        try:
            self.tui.alreadyRegistered = rhnreg.reserveUser(self.userNameEntry,
                                                            self.passwordEntry)
        except up2dateErrors.ValidationError, e:
            FatalErrorAbort("The server indicated an error:\n" + e.errmsg ) ;
            self.g.setCurrent(self.userNameEntry)
            return 0
        except up2dateErrors.CommunicationError,e:
            FatalErrorAbort("There was an error communicating with the registration server:\n" + e.errmsg)

        if not rhnreg.validateEmail(self.emailEntry):
            FatalErrorAbort("You must enter a valid e-mail address."  ) ;
            return 0
        
        return 1


    def saveResults(self):
        self.tui.userName = self.userNameEntry
        self.tui.password = self.passwordEntry
        self.tui.email = self.emailEntry
        
        
    def run(self):
        valid = 0
        valid = self.validateFields()
	if not valid:
	    FatalErrorAbort("exit") ;
        return "next"
    

class HardwareWindow:

    def __init__(self, tui):
        self.name = "HardwareWindow"

        self.tui = tui

        self.profileEntry = swcfg.GetProfileName() ;
        tui.version = up2dateUtils.getVersion()

    def saveResults(self):
        self.tui.profileName = self.profileEntry

    def run(self):
        return "next"

class PackagesWindow:

    def __init__(self, tui):
        self.name = "PackagesWindow"

        self.tui = tui



    def setScale(self, amount, total):
	pass
        

    def saveResults(self):
	return

        
    def run(self):
        return "next"

class SendWindow:

    def __init__(self, tui):
        self.tui = tui
        self.name = "SendWindow"
        

    def saveResults(self):
        pass


    def run(self):
        return "next"

class FinishWindow:

    def __init__(self, tui):
        self.name = "FinishWindow"

        self.tui = tui

        # send the data
        try:
            rhnreg.registerUser(tui.userName, tui.password, tui.email)
        except up2dateErrors.CommunicationError, e:
            FatalErrorAbort("Problem registering user name:\n" + e.errmsg)
        except:
            FatalErrorAbort("Problem registering username.")


        try:
            systemId = rhnreg.registerSystem(tui.userName, tui.password,
                                             tui.profileName)
        except up2dateErrors.CommunicationError, e:
            FatalErrorAbort("Problem registering system:\n" + e.errmsg)
        except up2dateErrors.RhnUuidUniquenessError, e:
            FatalErrorAbort("Problem registering system:\n" + e.errmsg)
        except:
            FatalErrorAbort("Problem registering system.")

        # write the system id out.
        if not rhnreg.writeSystemId(systemId):
            FatalErrorAbort("Problem writing out system id to disk.")

        # include the info from the oeminfo file as well
        self.oemInfo = rhnreg.getOemInfo()
        
        # dont send if already registered, do send if they have oemInfo
        if ( not self.tui.alreadyRegistered ) or ( len(self.oemInfo)):
            # send product registration information
            if rhnreg.cfg['supportsUpdateContactInfo']:
                #rhnreg.updateContactInfo(self.username, self.password,  self.tui.productInfo,self.tui.oemInfo)
                try:
                    rhnreg.updateContactInfo(tui.userName, tui.password,  self.tui.productInfo)
                except up2dateErrors.CommunicationError, e:
                    FatalErrorAbort("Problem registering personal information:\n" + e.errmsg)
                except:
                    print sys.exc_info()
                    print sys.exc_type
                    
                    FatalErrorAbort("Problem registering personal information")
                    
            else:
                rhnreg.registerProduct(systemId, self.tui.productInfo,self.tui.oemInfo)
                try:
                    rhnreg.registerProduct(systemId, self.tui.productInfo,self.tui.oemInfo)
                except up2dateErrors.CommunicationError, e:
                    FatalErrorAbort("Problem registering personal information:\n" + e.errmsg)
                except:
                    FatalErrorAbort("Problem registering personal information")

	tui.includeHardware = 0
        if tui.includeHardware:
            try:
                rhnreg.sendHardware(systemId, tui.hardware)
            except up2dateErrors.CommunicationError, e:
                FatalErrorAbort("Problem sending hardware profile:\n" + e.errmsg)
            except:
                FatalErrorAbort("Problem sending hardware profile.")

        # build up package list if necessary
	tui.includePackages = 0
        if tui.includePackages:
            try:
                rhnreg.sendPackages(systemId, tui.selectedPackages)
            except up2dateErrors.CommunicationError, e:
                FatalErrorAbort("Problem sending package list:\n" + e.errmsg)
            except:
                FatalErrorAbort("Problem sending package list.")

        rhnreg.startRhnsd()


    def setScale(self, amount, total):
	pass

    def saveResults(self):
        pass


    def run(self):
        result = self.g.runOnce()

        return "next"


class Tui:

    def __init__(self ):
        self.alreadyRegistered = 0
	self.drawFrame()
	
        self.windows = [
            StartWindow,
            PrivacyWindow,
            InfoWindow,
            HardwareWindow,
            PackagesWindow,
            SendWindow,
            FinishWindow
            ]
       
    def __del__(self):
#        self.screen.finish()
	return


    def drawFrame(self):
	return

    def initResults(self):
        self.userName = ""
        self.password = ""
        self.enableEmail = 1
        self.email = ""

        self.oemInfo = {}
        self.productInfo = {
            "reg_num" : "",
            "title" : "",
            "first_name" : "",
            "last_name" : "",
            "company" : "",
            "position" : "",
            "address1" : "",
            "address2" : "",
            "city" : "",
            "state" : "",
            "zip" : "",
            "country" : "",
            "phone" : "",
            "fax" : "",
            "contact_email" : 1,
            "contact_mail" : 0,
            "contact_phone" : 0,
            "contact_fax" : 0,
            "contact_special" : 0,
            "contact_newsletter" : 0
            }
        
        self.profileName = ""
        self.includeHardware = 1
        self.version = ""
        self.hostname = ""
        self.ip = ""
        self.cpu = ""
        self.speed = 0
        self.ram = ""
        
        self.includePackages = 0
        self.packageList = []
        self.selectedPackages = []

        
    def run(self):
        self.initResults()
        
        try:
            index = 0
            while index < len(self.windows):
                win = self.windows[index](self)
                # dont get all the gory user info if they already registered
                if win.name == "ProductWindow" and self.alreadyRegistered:
                    index = index + 1
                    continue
               
                result = win.run()
                win.saveResults()

                if result == "back":
                    # hardware window
                    if win == self.windows[4] and self.alreadyRegistered:
                        index = index - 2
                    else:
                        index = index - 1
                elif result == "next":
                    # info window
                    if win == self.windows[2] and self.alreadyRegistered:
                        index = index + 2
                    else:
                        index = index + 1
                elif result == "cancel":
                    return
	finally:
	   return 

def main():
    test = 0    

    if len(sys.argv) > 1:
        if sys.argv[1] == "-t" or sys.argv[1] == "--test":
            test = 1
            
    if not rpcServer.hasSSL():
        FatalErrorAbort("SSL Warning: Your system libraries do not support SSL (secure) connections.  Any data that you send or receive from Red Hat Network will be transmitted in the clear. Continue anyway? " )
       

#    if geteuid() != 0 and not test:
#        FatalErrorAbort("You must run the RHN registration program as root.")


    if rhnreg.registered() and not test:
        FatalErrorAbort( "Already Registered. This system appears to already be registered.  Continue anyway?" ) 
        
    tui = Tui()
    tui.run()

    
if __name__ == "__main__":
    main()
